@@ -302,38 +302,68 @@ def _remove_failed_setup_state_from_session(item):
         setup_state.stack = []
 
 
-def _should_hard_fail_on_error(session_config, report):
+def _get_rerun_filter_regex(item, regex_name):
+    rerun_marker = _get_marker(item)
+
+    if rerun_marker is not None and regex_name in rerun_marker.kwargs:
+        regex = rerun_marker.kwargs[regex_name]
+        if isinstance(regex, str):
+            regex = [regex]
+    else:
+        regex = getattr(item.session.config.option, regex_name)
+
+    return regex
+
+
+def _matches_any_rerun_error(rerun_errors, report):
+    for rerun_regex in rerun_errors:
+        try:
+            if re.search(rerun_regex, report.longrepr.reprcrash.message):
+                return True
+        except AttributeError:
+            if re.search(rerun_regex, report.longreprtext):
+                return True
+    return False
+
+
+def _matches_any_rerun_except_error(rerun_except_errors, report):
+    for rerun_regex in rerun_except_errors:
+        if re.search(rerun_regex, report.longrepr.reprcrash.message):
+            return True
+    return False
+
+
+def _should_hard_fail_on_error(item, report):
     if report.outcome != "failed":
         return False
 
-    rerun_errors = session_config.option.only_rerun
-    rerun_except_errors = session_config.option.rerun_except
-
-    if not rerun_errors and not rerun_except_errors:
+    rerun_errors = _get_rerun_filter_regex(item, "only_rerun")
+    rerun_except_errors = _get_rerun_filter_regex(item, "rerun_except")
 
+    if (not rerun_errors) and (not rerun_except_errors):
+        # Using neither --only-rerun nor --rerun-except
         return False
 
-    if rerun_errors:
-        for rerun_regex in rerun_errors:
-            try:
-                if re.search(rerun_regex, report.longrepr.reprcrash.message):
-                    return False
-            except AttributeError:
-                if re.search(rerun_regex, report.longreprtext):
-                    return False
+    elif rerun_errors and (not rerun_except_errors):
+        # Using --only-rerun but not --rerun-except
+        return not _matches_any_rerun_error(rerun_errors, report)
 
-    if rerun_except_errors:
-        for rerun_regex in rerun_except_errors:
-            if not re.search(rerun_regex, report.longrepr.reprcrash.message):
+    elif (not rerun_errors) and rerun_except_errors:
+        # Using --rerun-except but not --only-rerun
+        return _matches_any_rerun_except_error(rerun_except_errors, report)
 
-                return False
-
-    return True
+    else:
+        # Using both --only-rerun and --rerun-except
+        matches_rerun_only = _matches_any_rerun_error(rerun_errors, report)
+        matches_rerun_except = _matches_any_rerun_except_error(
+            rerun_except_errors, report
+        )
+        return (not matches_rerun_only) or matches_rerun_except
 
 
 def _should_not_rerun(item, report, reruns):
     xfail = hasattr(report, "wasxfail")
-    is_terminal_error = _should_hard_fail_on_error(item.session.config, report)
+    is_terminal_error = _should_hard_fail_on_error(item, report)
     condition = get_reruns_condition(item)
     return (
         item.execution_count > reruns
